import asyncio
import inspect
import json
import time
import uuid
from functools import wraps
from typing import Callable, Dict, List, Optional, TypedDict, Union

from chainlo.config import config
from chainlo.context import context, local_steps
from chainlo.data import get_data_layer
from chainlo.element import Element
from chainlo.logger import logger
from chainlo.telemetry import trace_event
from chainlo.types import FeedbackDict
from literalai import BaseGeneration
from literalai.helper import utc_now
from literalai.step import StepType, TrueStepType


class StepDict(TypedDict, total=False):
    name: str
    type: StepType
    id: str
    threadId: str
    parentId: Optional[str]
    disableFeedback: bool
    streaming: bool
    waitForAnswer: Optional[bool]
    isError: Optional[bool]
    metadata: Dict
    input: str
    output: str
    createdAt: Optional[str]
    start: Optional[str]
    end: Optional[str]
    generation: Optional[Dict]
    showInput: Optional[Union[bool, str]]
    language: Optional[str]
    indent: Optional[int]
    feedback: Optional[FeedbackDict]


def step(
    original_function: Optional[Callable] = None,
    *,
    name: Optional[str] = "",
    type: TrueStepType = "undefined",
    id: Optional[str] = None,
    disable_feedback: bool = True,
    root: bool = False,
    language: Optional[str] = None,
    show_input: Union[bool, str] = False,
):
    """Step decorator for async and sync functions."""

    def wrapper(func: Callable):
        nonlocal name
        if not name:
            name = func.__name__

        # Handle async decorator

        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                async with Step(
                    type=type,
                    name=name,
                    id=id,
                    disable_feedback=disable_feedback,
                    root=root,
                    language=language,
                    show_input=show_input,
                ) as step:
                    try:
                        step.input = {"args": args, "kwargs": kwargs}
                    except:
                        pass
                    result = await func(*args, **kwargs)
                    try:
                        if result and not step.output:
                            step.output = result
                    except:
                        pass
                    return result

            return async_wrapper
        else:
            # Handle sync decorator
            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                with Step(
                    type=type,
                    name=name,
                    id=id,
                    disable_feedback=disable_feedback,
                    root=root,
                    language=language,
                    show_input=show_input,
                ) as step:
                    try:
                        step.input = {"args": args, "kwargs": kwargs}
                    except:
                        pass
                    result = func(*args, **kwargs)
                    try:
                        if result and not step.output:
                            step.output = result
                    except:
                        pass
                    return result

            return sync_wrapper

    func = original_function
    if not func:
        return wrapper
    else:
        return wrapper(func)


class Step:
    # Constructor
    name: str
    type: TrueStepType
    id: str
    parent_id: Optional[str]
    disable_feedback: bool

    streaming: bool
    persisted: bool

    root: bool
    show_input: Union[bool, str]

    is_error: Optional[bool]
    metadata: Dict
    thread_id: str
    created_at: Union[str, None]
    start: Union[str, None]
    end: Union[str, None]
    generation: Optional[BaseGeneration]
    language: Optional[str]
    elements: Optional[List[Element]]
    fail_on_persist_error: bool

    def __init__(
        self,
        name: Optional[str] = config.ui.name,
        type: TrueStepType = "undefined",
        id: Optional[str] = None,
        parent_id: Optional[str] = None,
        elements: Optional[List[Element]] = None,
        disable_feedback: bool = True,
        root: bool = False,
        language: Optional[str] = None,
        show_input: Union[bool, str] = False,
    ):
        trace_event(f"init {self.__class__.__name__} {type}")
        time.sleep(0.001)
        self._input = ""
        self._output = ""
        self.thread_id = context.session.thread_id
        self.name = name or ""
        self.type = type
        self.id = id or str(uuid.uuid4())
        self.disable_feedback = disable_feedback
        self.metadata = {}
        self.is_error = False
        self.show_input = show_input
        self.parent_id = parent_id
        self.root = root

        self.language = language
        self.generation = None
        self.elements = elements or []

        self.created_at = utc_now()
        self.start = None
        self.end = None

        self.streaming = False
        self.persisted = False
        self.fail_on_persist_error = False

    def _process_content(self, content, set_language=False):
        if content is None:
            return ""
        if isinstance(content, dict):
            try:
                processed_content = json.dumps(content, indent=4, ensure_ascii=False)
                if set_language:
                    self.language = "json"
            except TypeError:
                processed_content = str(content)
                if set_language:
                    self.language = "text"
        elif isinstance(content, str):
            processed_content = content
        else:
            processed_content = str(content)
            if set_language:
                self.language = "text"
        return processed_content

    @property
    def input(self):
        return self._input

    @input.setter
    def input(self, content: Union[Dict, str]):
        self._input = self._process_content(content, set_language=False)

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, content: Union[Dict, str]):
        self._output = self._process_content(content, set_language=True)

    def to_dict(self) -> StepDict:
        _dict: StepDict = {
            "name": self.name,
            "type": self.type,
            "id": self.id,
            "threadId": self.thread_id,
            "parentId": self.parent_id,
            "disableFeedback": self.disable_feedback,
            "streaming": self.streaming,
            "metadata": self.metadata,
            "input": self.input,
            "isError": self.is_error,
            "output": self.output,
            "createdAt": self.created_at,
            "start": self.start,
            "end": self.end,
            "language": self.language,
            "showInput": self.show_input,
            "generation": self.generation.to_dict() if self.generation else None,
        }
        return _dict

    async def update(self):
        """
        Update a step already sent to the UI.
        """
        trace_event("update_step")

        if self.streaming:
            self.streaming = False

        step_dict = self.to_dict()
        data_layer = get_data_layer()

        if data_layer:
            try:
                asyncio.create_task(data_layer.update_step(step_dict.copy()))
            except Exception as e:
                if self.fail_on_persist_error:
                    raise e
                logger.error(f"Failed to persist step update: {str(e)}")

        tasks = [el.send(for_id=self.id) for el in self.elements]
        await asyncio.gather(*tasks)

        if config.ui.hide_cot and self.parent_id:
            return

        if not config.features.prompt_playground and "generation" in step_dict:
            step_dict.pop("generation", None)

        await context.emitter.update_step(step_dict)

        return True

    async def remove(self):
        """
        Remove a step already sent to the UI.
        """
        trace_event("remove_step")

        step_dict = self.to_dict()
        data_layer = get_data_layer()

        if data_layer:
            try:
                asyncio.create_task(data_layer.delete_step(self.id))
            except Exception as e:
                if self.fail_on_persist_error:
                    raise e
                logger.error(f"Failed to persist step deletion: {str(e)}")

        await context.emitter.delete_step(step_dict)

        return True

    async def send(self):
        if self.persisted:
            return

        if config.code.author_rename:
            self.name = await config.code.author_rename(self.name)

        if self.streaming:
            self.streaming = False

        step_dict = self.to_dict()

        data_layer = get_data_layer()

        if data_layer:
            try:
                asyncio.create_task(data_layer.create_step(step_dict.copy()))
                self.persisted = True
            except Exception as e:
                if self.fail_on_persist_error:
                    raise e
                logger.error(f"Failed to persist step creation: {str(e)}")

        tasks = [el.send(for_id=self.id) for el in self.elements]
        await asyncio.gather(*tasks)

        if config.ui.hide_cot and self.parent_id:
            return self.id

        if not config.features.prompt_playground and "generation" in step_dict:
            step_dict.pop("generation", None)

        await context.emitter.send_step(step_dict)

        return self.id

    async def stream_token(self, token: str, is_sequence=False):
        """
        Sends a token to the UI.
        Once all tokens have been streamed, call .send() to end the stream and persist the step if persistence is enabled.
        """

        if not self.streaming:
            self.streaming = True
            step_dict = self.to_dict()
            await context.emitter.stream_start(step_dict)

        if is_sequence:
            self.output = token
        else:
            self.output += token

        assert self.id

        if config.ui.hide_cot and self.parent_id:
            return

        await context.emitter.send_token(
            id=self.id, token=token, is_sequence=is_sequence
        )

    # Handle parameter less decorator
    def __call__(self, func):
        return step(
            original_function=func,
            type=self.type,
            name=self.name,
            id=self.id,
            parent_id=self.parent_id,
            thread_id=self.thread_id,
            disable_feedback=self.disable_feedback,
        )

    # Handle Context Manager Protocol
    async def __aenter__(self):
        self.start = utc_now()
        previous_steps = local_steps.get() or []
        parent_step = previous_steps[-1] if previous_steps else None

        if not self.parent_id and not self.root:
            if parent_step:
                self.parent_id = parent_step.id
            elif context.session.root_message:
                self.parent_id = context.session.root_message.id
        context.active_steps.append(self)
        local_steps.set(previous_steps + [self])
        await self.send()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.end = utc_now()

        if self in context.active_steps:
            context.active_steps.remove(self)

        local_active_steps = local_steps.get()
        if local_active_steps and self in local_active_steps:
            local_active_steps.remove(self)
            local_steps.set(local_active_steps)

        await self.update()

    def __enter__(self):
        self.start = utc_now()

        previous_steps = local_steps.get() or []
        parent_step = previous_steps[-1] if previous_steps else None

        if not self.parent_id and not self.root:
            if parent_step:
                self.parent_id = parent_step.id
            elif context.session.root_message:
                self.parent_id = context.session.root_message.id
        context.active_steps.append(self)
        local_steps.set(previous_steps + [self])

        asyncio.create_task(self.send())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = utc_now()
        if self in context.active_steps:
            context.active_steps.remove(self)

        local_active_steps = local_steps.get()
        if local_active_steps and self in local_active_steps:
            local_active_steps.remove(self)
            local_steps.set(local_active_steps)

        asyncio.create_task(self.update())



import os
import re
import shutil
import sys
from typing import Union
from pathlib import Path

import click
import questionary
import toml

from chainlo.agenta_cli.agenta.cli import helper
from chainlo.agenta_cli.agenta.cli import variant_configs
from chainlo.agenta_cli.agenta.cli import variant_commands
from chainlo.agenta_cli.agenta.cli import evaluation_commands

from chainlo.agenta_cli.agenta.client.backend.client import AgentaApi

BACKEND_URL_SUFFIX = os.environ.get("BACKEND_URL_SUFFIX", "api")


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    try:
        try:
            from importlib.metadata import PackageNotFoundError, version
        except ImportError:
            from importlib_metadata import PackageNotFoundError, version
        package_version = version("agenta")
    except PackageNotFoundError:
        package_version = "package is not installed"
    click.echo(f"Agenta CLI version: {package_version}")
    ctx.exit()


def check_latest_version() -> Union[str, None]:
    import requests

    try:
        response = requests.get("https://pypi.org/pypi/agenta/json", timeout=360)
        response.raise_for_status()
        latest_version = response.json()["info"]["version"]
        return latest_version
    except (requests.RequestException, KeyError):
        return None


def notify_update(available_version: str):
    import pkg_resources

    # installed_version = pkg_resources.get_distribution("agenta").version
    # if available_version > installed_version:
    #     click.echo(
    #         click.style(
    #             f"A new release of agenta is available: {installed_version} → {available_version}",
    #             fg="yellow",
    #         )
    #     )
    #     click.echo(
    #         click.style("To upgrade, run: pip install --upgrade agenta", fg="yellow")
    #     )


@click.group()
@click.option(
    "--version",
    "-v",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
)
def cli():
    latest_version = check_latest_version()
    # if latest_version:
    #     notify_update(latest_version)


@click.command()
@click.option("--app_name", default="")
@click.option("--backend_host", default="")
def init(app_name: str, backend_host: str):
    init_option = "Blank App" if backend_host != "" and app_name != "" else ""
    """Initialize a new Agenta app with the template files."""

    api_key = os.getenv("AGENTA_API_KEY")

    if not app_name:
        while True:
            app_name = questionary.text("Please enter the app name").ask()
            if app_name and re.match("^[a-zA-Z0-9_-]+$", app_name):
                break
            else:
                if app_name is None:  # User pressed Ctrl+C
                    sys.exit(0)
                else:
                    print(
                        "Invalid input. Please use only alphanumeric characters without spaces."
                    )

    try:
        backend_hosts = {
            "https://cloud.agenta.ai": "On agenta cloud",
            "http://localhost": "On my local machine",
        }
        where_question = backend_hosts.get(backend_host, "On a remote machine")
        if not backend_host:
            where_question = questionary.select(
                "Where are you running agenta?",
                choices=[
                    "On agenta cloud",
                    "On my local machine",
                    "On a remote machine",
                ],
            ).ask()

            if where_question == "On my local machine":
                backend_host = "http://localhost"
            elif where_question == "On a remote machine":
                backend_host = questionary.text(
                    "Please provide the IP or URL of your remote host"
                ).ask()
            elif where_question == "On agenta cloud":
                global_backend_host = helper.get_global_config("host")
                if global_backend_host:
                    backend_host = global_backend_host
                else:
                    backend_host = "https://cloud.agenta.ai"

                if not api_key:
                    api_key = helper.get_api_key(backend_host)

            elif where_question is None:  # User pressed Ctrl+C
                sys.exit(0)
        backend_host = (
            backend_host
            if backend_host.startswith("http://") or backend_host.startswith("https://")
            else "http://" + backend_host
        )

        # initialize the client with the backend url and api key
        client = AgentaApi(
            base_url=f"{backend_host}/{BACKEND_URL_SUFFIX}",
            api_key=api_key if where_question == "On agenta cloud" else "",
        )

        # list of user organizations
        user_organizations = []

        # validate the api key if it is provided
        if where_question == "On agenta cloud":
            try:
                key_prefix = api_key.split(".")[0]
                client.validate_api_key(key_prefix=key_prefix)

                # Make request to fetch user organizations after api key validation
                organizations = client.list_organizations()
                if len(organizations) >= 1:
                    user_organizations = organizations
            except Exception as ex:
                if ex.status_code == 401:
                    click.echo(click.style("Error: Invalid API key", fg="red"))
                    sys.exit(1)
                else:
                    click.echo(click.style(f"Error: {ex}", fg="red"))
                    sys.exit(1)

        filtered_org = None
        if where_question == "On agenta cloud":
            which_organization = questionary.select(
                "Which organization do you want to create the app for?",
                choices=[
                    f"{org.name}: {org.description}" for org in user_organizations
                ],
            ).ask()
            filtered_org = next(
                (
                    org
                    for org in user_organizations
                    if org.name == which_organization.split(":")[0]
                ),
                None,
            )

        # Get app_id after creating new app in the backend server
        try:
            app_id = client.create_app(
                app_name=app_name,
                organization_id=filtered_org.id if filtered_org else None,
            ).app_id
        except Exception as ex:
            click.echo(click.style(f"Error: {ex}", fg="red"))
            sys.exit(1)

        # Set app toml configuration
        config = {
            "app_name": app_name,
            "app_id": app_id,
            "backend_host": backend_host,
            "api_key": api_key if where_question == "On agenta cloud" else None,
        }
        with open("config.toml", "w") as config_file:
            toml.dump(config, config_file)

        # Ask for init option
        if not init_option:
            init_option = questionary.select(
                "How do you want to initialize your app?",
                choices=["Blank App", "Start from template"],
            ).ask()

            # If the user selected the second option, show a list of available templates
            if init_option == "Start from template":
                current_dir = Path.cwd()
                template_dir = Path(__file__).parent.parent / "templates"
                templates = [
                    folder.name for folder in template_dir.iterdir() if folder.is_dir()
                ]
                template_desc = [
                    toml.load((template_dir / name / "template.toml"))["short_desc"]
                    for name in templates
                ]

                # Show the templates to the user
                template = questionary.select(
                    "Which template do you want to use?",
                    choices=[
                        questionary.Choice(
                            title=f"{template} - {template_desc}", value=template
                        )
                        for template, template_desc in zip(templates, template_desc)
                    ],
                ).ask()

                # Copy the template files to the current directory
                chosen_template_dir = template_dir / template
                for file in chosen_template_dir.glob("*"):
                    if file.name != "template.toml" and not file.is_dir():
                        shutil.copy(file, current_dir / file.name)
            elif init_option is None:  # User pressed Ctrl+C
                sys.exit(0)

        # Create a .gitignore file and add some default environment folder names to it
        gitignore_content = (
            "# Environments \nenv/\nvenv/\nENV/\nenv.bak/\nvenv.bak/\nmyenv/\n"
        )
        with open(".gitignore", "w") as gitignore_file:
            gitignore_file.write(gitignore_content)

        click.echo("App initialized successfully")
        if init_option == "Start from template":
            click.echo(
                "Please check the README.md for further instructions to setup the template."
            )
    except Exception as ex:
        click.echo(click.style(f"Error: {ex}", fg="red"))
        sys.exit(1)


# Add the commands to the CLI group
cli.add_command(init)
cli.add_command(variant_configs.config)
cli.add_command(variant_commands.variant)
cli.add_command(evaluation_commands.evaluation)

if __name__ == "__main__":
    cli()



import click
from chainlo.agenta_cli.agenta.client import client


@click.group()
def evaluation():
    """Commands for evaluations."""
    pass


# TODO: Remove hardcoded values
@evaluation.command(name="run")
def run_evaluation_cli():
    """Run an evaluation."""

    try:
        client.run_evaluation(
            app_name="sss",
            host="http://localhost",
        )
    except Exception as ex:
        click.echo(click.style(f"Error while running evaluation: {ex}", fg="red"))



from chainlo.agenta_cli.agenta.client.exceptions import APIRequestError
from chainlo.agenta_cli.agenta.client.backend.client import AgentaApi
import os
import logging
from typing import Any, Optional

from .utils.globals import set_global

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


BACKEND_URL_SUFFIX = os.environ.get("BACKEND_URL_SUFFIX", "api")
CLIENT_API_KEY = os.environ.get("AGENTA_API_KEY")
CLIENT_HOST = os.environ.get("AGENTA_HOST", "http://localhost")

# initialize the client with the backend url and api key
backend_url = f"{CLIENT_HOST}/{BACKEND_URL_SUFFIX}"
client = AgentaApi(
    base_url=backend_url,
    api_key=CLIENT_API_KEY if CLIENT_API_KEY else "",
)


class AgentaSingleton:
    """Singleton class to save all the "global variables" for the sdk."""

    _instance = None
    setup = None
    config = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(AgentaSingleton, cls).__new__(cls)
        return cls._instance

    def init(
        self,
        app_name: Optional[str] = None,
        base_name: Optional[str] = None,
        api_key: Optional[str] = None,
        base_id: Optional[str] = None,
        host: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Main function to initialize the singleton.

        Initializes the singleton with the given `app_name`, `base_name`, and `host`. If any of these arguments are not provided,
        the function will look for them in environment variables.

        Args:
            app_name (Optional[str]): Name of the Agenta application. Defaults to None. If not provided, will look for "AGENTA_APP_NAME" in environment variables.
            base_name (Optional[str]): Base name for the Agenta setup. Defaults to None. If not provided, will look for "AGENTA_BASE_NAME" in environment variables.
            host (Optional[str]): Host name of the backend server. Defaults to None. If not provided, will look for "AGENTA_HOST" in environment variables.
            kwargs (Any): Additional keyword arguments.

        Raises:
            ValueError: If `app_name`, `base_name`, or `host` are not specified either as arguments or in the environment variables.
        """
        if app_name is None:
            app_name = os.environ.get("AGENTA_APP_NAME")
        if base_name is None:
            base_name = os.environ.get("AGENTA_BASE_NAME")
        if api_key is None:
            api_key = os.environ.get("AGENTA_API_KEY")
        if base_id is None:
            base_id = os.environ.get("AGENTA_BASE_ID")
        if host is None:
            host = os.environ.get("AGENTA_HOST", "http://localhost")

        if base_id is None:
            if app_name is None or base_name is None:
                print(
                    f"Warning: Your configuration will not be saved permanently since app_name and base_name are not provided."
                )
            else:
                try:
                    apps = client.list_apps(app_name=app_name)
                    if len(apps) == 0:
                        raise APIRequestError(f"App with name {app_name} not found")

                    app_id = apps[0].app_id
                    if not app_id:
                        raise APIRequestError(
                            f"App with name {app_name} does not exist on the server."
                        )

                    bases = client.list_bases(app_id=app_id, base_name=base_name)
                    if len(bases) == 0:
                        raise APIRequestError(f"No base was found for the app {app_id}")

                    base_id = bases[0].base_id
                except Exception as ex:
                    raise APIRequestError(
                        f"Failed to get base id and/or app_id from the server with error: {ex}"
                    )
        self.base_id = base_id
        self.host = host
        self.api_key = api_key
        self.config = Config(base_id=base_id, host=host)


class Config:
    def __init__(self, base_id, host):
        self.base_id = base_id
        self.host = host
        if base_id is None or host is None:
            self.persist = False
        else:
            self.persist = True

    def register_default(self, overwrite=False, **kwargs):
        """alias for default"""
        return self.default(overwrite=overwrite, **kwargs)

    def default(self, overwrite=False, **kwargs):
        """Saves the default parameters to the app_name and base_name in case they are not already saved.
        Args:
            overwrite: Whether to overwrite the existing configuration or not
            **kwargs: A dict containing the parameters
        """
        self.set(
            **kwargs
        )  # In case there is no connectivity, we still can use the default values
        try:
            self.push(config_name="default", overwrite=overwrite, **kwargs)
        except Exception as ex:
            logger.warning(
                "Unable to push the default configuration to the server." + str(ex)
            )

    def push(self, config_name: str, overwrite=True, **kwargs):
        """Pushes the parameters for the app variant to the server
        Args:
            config_name: Name of the configuration to push to
            overwrite: Whether to overwrite the existing configuration or not
            **kwargs: A dict containing the parameters
        """
        if not self.persist:
            return
        try:
            client.save_config(
                base_id=self.base_id,
                config_name=config_name,
                parameters=kwargs,
                overwrite=overwrite,
            )
        except Exception as ex:
            logger.warning(
                "Failed to push the configuration to the server with error: " + str(ex)
            )

    def pull(self, config_name: str = "default", environment_name: str = None):
        """Pulls the parameters for the app variant from the server and sets them to the config"""
        if not self.persist and (
            config_name != "default" or environment_name is not None
        ):
            raise Exception(
                "Cannot pull the configuration from the server since the app_name and base_name are not provided."
            )
        if self.persist:
            try:
                if environment_name:
                    config = client.get_config(
                        base_id=self.base_id, environment_name=environment_name
                    )

                else:
                    config = client.get_config(
                        base_id=self.base_id,
                        config_name=config_name,
                    )
            except Exception as ex:
                logger.warning(
                    "Failed to pull the configuration from the server with error: "
                    + str(ex)
                )
        try:
            self.set(**config.parameters)
        except Exception as ex:
            logger.warning("Failed to set the configuration with error: " + str(ex))

    def all(self):
        """Returns all the parameters for the app variant"""
        return {
            k: v
            for k, v in self.__dict__.items()
            if k
            not in ["app_name", "base_name", "host", "base_id", "api_key", "persist"]
        }

    # function to set the parameters for the app variant
    def set(self, **kwargs):
        """Sets the parameters for the app variant

        Args:
            **kwargs: A dict containing the parameters
        """
        for key, value in kwargs.items():
            setattr(self, key, value)


def init(app_name=None, base_name=None, **kwargs):
    """Main function to be called by the user to initialize the sdk.

    Args:
        app_name: _description_. Defaults to None.
        base_name: _description_. Defaults to None.
    """
    singleton = AgentaSingleton()
    singleton.init(app_name=app_name, base_name=base_name, **kwargs)
    set_global(setup=singleton.setup, config=singleton.config)
